from setuptools import setup

import kindling


setup(
  name='kindling',
  version=kindling.__version__,
  description='Basic building blocks for pytorch.',
  author='Samuel Ainsworth',
  author_email='skainsworth@gmail.com',
  url='https://github.com/samuela/kindling',
  packages=['kindling'],
  install_requires=[]
)
